﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2013 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Encodings.IntegerVectorEncoding;
using HeuristicLab.Encodings.SymbolicExpressionTreeEncoding;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Random;

namespace HeuristicLab.Problems.GrammaticalEvolution {
  /// <summary>
  /// DepthFirstMapper
  /// </summary>
  [Item("DepthFirstMapper", "")]
  [StorableClass]
  public class DepthFirstMapper : GenotypeToPhenotypeMapper {

    [StorableConstructor]
    protected DepthFirstMapper(bool deserializing) : base(deserializing) { }
    protected DepthFirstMapper(DepthFirstMapper original, Cloner cloner) : base(original, cloner) { }
    public DepthFirstMapper() : base() { }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new DepthFirstMapper(this, cloner);
    }


    /// <summary>
    /// Maps a genotype (an integer vector) to a phenotype (a symbolic expression tree).
    /// Depth-first approach.
    /// </summary>
    /// <param name="grammar">grammar definition</param>
    /// <param name="genotype">integer vector, which should be mapped to a tree</param>
    /// <returns>phenotype (a symbolic expression tree)</returns>
    public override SymbolicExpressionTree Map(ISymbolicExpressionGrammar grammar,
                                               IntegerVector genotype) {

      SymbolicExpressionTree tree = new SymbolicExpressionTree();
      var rootNode = (SymbolicExpressionTreeTopLevelNode)grammar.ProgramRootSymbol.CreateTreeNode();
      if (rootNode.HasLocalParameters) rootNode.ResetLocalParameters(new MersenneTwister());
      var startNode = (SymbolicExpressionTreeTopLevelNode)grammar.StartSymbol.CreateTreeNode();
      if (startNode.HasLocalParameters) startNode.ResetLocalParameters(new MersenneTwister());
      rootNode.AddSubtree(startNode);
      tree.Root = rootNode;

      int genotypeIndex = 0;
      int currSubtreeCount = 1;

      MapDepthFirstRecursively(startNode, genotype,
                               grammar, genotype.Length,
                               ref genotypeIndex, ref currSubtreeCount);

      return tree;
    }


    /// <summary>
    /// Genotype-to-Phenotype mapper (recursive depth-first approach).
    /// Appends maximum allowed children (non-terminal symbols) to
    /// <paramref name="currentNode"/>, as long as <paramref name="currSubtreeCount"/>
    /// doesn't exceed <paramref name="maxSubtreeCount"/>.
    /// If at most <paramref name="maxSubtreeCount"/> subtrees were created,
    /// each non-full node is filled with randomly chosen nodes
    /// (non-terminal and terminal), and each non-terminal node is again filled with a terminal node.
    /// </summary>
    /// <param name="currentNode">current parent node</param>
    /// <param name="genotype">integer vector, which should be mapped to a tree</param>
    /// <param name="grammar">grammar definition to determine the allowed child symbols for currentNode </param>
    /// <param name="maxSubtreeCount">maximum allowed subtrees (= number of used genomes)</param>
    /// <param name="genotypeIndex">current index in integer vector</param>
    /// <param name="currSubtreeCount">number of already determined subtrees (filled or still incomplete)</param>
    private void MapDepthFirstRecursively(ISymbolicExpressionTreeNode currentNode,
                                          IntegerVector genotype,
                                          ISymbolicExpressionGrammar grammar,
                                          int maxSubtreeCount,
                                          ref int genotypeIndex,
                                          ref int currSubtreeCount) {
      if (currSubtreeCount < maxSubtreeCount) {

        var newNode = GetNewChildNode(currentNode, genotype, grammar, genotypeIndex);

        if ((currSubtreeCount + newNode.Symbol.MinimumArity) > maxSubtreeCount) {
          // TODO: maybe check, if there is any node, which fits in the tree yet
          currentNode.AddSubtree(GetRandomTerminalNode(currentNode, grammar));
        } else {
          currentNode.AddSubtree(newNode);
          genotypeIndex++;
          currSubtreeCount += newNode.Symbol.MinimumArity;

          while (newNode.Symbol.MinimumArity > newNode.SubtreeCount) {
            MapDepthFirstRecursively(newNode, genotype,
                                     grammar, maxSubtreeCount,
                                     ref genotypeIndex, ref currSubtreeCount);
          }
        }

      } else {
        while (currentNode.Symbol.MinimumArity > currentNode.SubtreeCount) {
          var newNode = GetNewChildNode(currentNode, genotype, grammar, genotypeIndex);
          currentNode.AddSubtree(newNode);
          genotypeIndex++;
          while (newNode.Symbol.MinimumArity > newNode.SubtreeCount) {
            newNode.AddSubtree(GetRandomTerminalNode(newNode, grammar));
          }
        }
      }
    }
  }
}